/*
 *  Project: LKH - lect3
 *  File: es4.c
 *  Date: 01/24/2006
 *  Author. R. Gioiosa
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/proc_fs.h>

#define MODULE_VERS "1.0"
#define MODULE_NAME "LKH-ES4"

#define DEBUG_ON

#ifdef DEBUG_ON
#define Dprintk(ARG...) printk(KERN_DEBUG MODULE_NAME ": " ARG)
#else
#define Dprintk(ARG...) 
#endif

#define N 32

static struct proc_dir_entry* lkh_pde;
static struct proc_dir_entry* entry;
static int es4_var = 0;
struct semaphore es4_sem;
static char buf[N];

static int es4_write(struct file* file,const char __user* buffer,unsigned long count,void* data)
{
  if(!count)
    {
      printk(KERN_ERR "Count is 0!!\n");
      return 0;
    }
  if(count>N)
    {
      Dprintk("Count is too large, set to %d\n",N-1);
      count = N-1;
    }
  if(copy_from_user(buf,buffer,count))
    {
      printk(KERN_ERR "Error while copying from user space!\n");
      return -EFAULT;
    }
  buf[count] = '\0';

  down(&es4_sem);
  if(sscanf(buf,"%d",&es4_var) != 1)
    {
      up(&es4_sem);
      printk(KERN_ERR "The value provided is not a valid number!\n");
      return -EFAULT;
    }
  Dprintk("writing es4_var=%d\n",es4_var);
  up(&es4_sem);

  return count;
}

static int es4_read(char* page,char** start,off_t off,int count,int* eof,void* data)
{
  int written = 0;

  down(&es4_sem);
  written = sprintf(page,"var = %d\n",es4_var);
  up(&es4_sem);
  *start = page+off;
  if(strlen(page) == written)
    *eof = 1;
  Dprintk("es4_read done.\n");
  return written-off;

}

static int __init es4_init(void)
{
  lkh_pde = proc_mkdir("lkh",NULL);
  if(!lkh_pde)
    {
      printk(KERN_ERR "%s: error creating proc_dir entry!\n",MODULE_NAME);
      goto err;
    }
  Dprintk("proc dir created\n");
  entry = create_proc_entry("foo",0644,lkh_pde);
  if(!entry)
    {
      printk(KERN_ERR "%s: error creating proc_entry!\n",MODULE_NAME);
      goto err_dir;
    }
  entry->data = NULL;
  entry->owner = THIS_MODULE;
  entry->read_proc = es4_read;
  entry->write_proc = es4_write;
  Dprintk("proc entry created\n");

  init_MUTEX(&es4_sem);

  Dprintk("module loaded\n");
  return 0;
 err_dir:
  remove_proc_entry("lkh",NULL);
  Dprintk("proc dir removed\n");
 err:
  return -1;
}

static void __exit es4_exit(void)
{
  remove_proc_entry("foo",lkh_pde);
  Dprintk("proc entry removed\n");
  remove_proc_entry("lkh",NULL);
  Dprintk("proc dir removed\n");
  Dprintk("module unloaded\n");
}

module_init(es4_init);
module_exit(es4_exit);

MODULE_AUTHOR("Roberto Gioiosa");
MODULE_DESCRIPTION("Linux Kernel Hacking 06 - Es4");
MODULE_LICENSE("GPL");
MODULE_VERSION(MODULE_VERS);
