/*
 *  Project: LKH - lect3
 *  File: es5.c
 *  Date: 01/24/2006
 *  Author. R. Gioiosa
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/proc_fs.h>

#define MODULE_VERS "1.0"
#define MODULE_NAME "LKH-ES5"

#define Dprintk(ARG...) printk(KERN_DEBUG MODULE_NAME ": " ARG)

#define N 32

static struct proc_dir_entry* lkh_pde;
static struct proc_dir_entry* entry;
static int es5_var = 0;
struct semaphore es5_sem;
static char buf[N];

static int debug = 0;
module_param(debug,int,S_IRUGO);

static int es5_write(struct file* file,const char __user* buffer,unsigned long count,void* data)
{
  if(!count)
    {
      printk(KERN_ERR "Count is 0!!\n");
      return 0;
    }
  if(count>N)
    {
      if(debug)
	Dprintk("Count is too large, set to %d\n",N-1);
      count = N-1;
    }
  if(copy_from_user(buf,buffer,count))
    {
      printk(KERN_ERR "Error while copying from user space!\n");
      return -EFAULT;
    }
  buf[count] = '\0';

  down(&es5_sem);
  if(sscanf(buf,"%d",&es5_var) != 1)
    {
      up(&es5_sem);
      printk(KERN_ERR "The value provided is not a valid number!\n");
      return -EFAULT;
    }
  up(&es5_sem);
  if(debug)
    Dprintk("writing es5_var=%d\n",es5_var);

  return count;
}

static int es5_read(char* page,char** start,off_t off,int count,int* eof,void* data)
{
  int written = 0;

  down(&es5_sem);
  written = sprintf(page,"var = %d\n",es5_var);
  up(&es5_sem);
  *start = page+off;
  if(strlen(page) == written)
    *eof = 1;
  if(debug)
    Dprintk("es5_read done.\n");
  return written-off;

}

static int __init es5_init(void)
{
  lkh_pde = proc_mkdir("lkh",NULL);
  if(!lkh_pde)
    {
      printk(KERN_ERR "%s: error creating proc_dir entry!\n",MODULE_NAME);
      goto err;
    }
  if(debug)
    Dprintk("proc dir created\n");
  entry = create_proc_entry("foo",0444,lkh_pde);
  if(!entry)
    {
      printk(KERN_ERR "%s: error creating proc_entry!\n",MODULE_NAME);
      goto err_dir;
    }
  entry->data = NULL;
  entry->owner = THIS_MODULE;
  entry->read_proc = es5_read;
  entry->write_proc = es5_write;
  if(debug)
    Dprintk("proc entry created\n");

  init_MUTEX(&es5_sem);

  if(debug)
    Dprintk("module loaded\n");
  return 0;
 err_dir:
  remove_proc_entry("lkh",NULL);
  if(debug)
    Dprintk("proc dir removed\n");
 err:
  return -1;
}

static void __exit es5_exit(void)
{
  remove_proc_entry("foo",lkh_pde);
  if(debug)
    Dprintk("proc entry removed\n");
  remove_proc_entry("lkh",NULL);
  if(debug)
    Dprintk("proc dir removed\n");
  if(debug)
    Dprintk("module unloaded\n");
}

module_init(es5_init);
module_exit(es5_exit);

MODULE_AUTHOR("Roberto Gioiosa");
MODULE_DESCRIPTION("Linux Kernel Hacking 06 - Es5");
MODULE_LICENSE("GPL");
MODULE_VERSION(MODULE_VERS);
