/*
 *  Project: LKH - lect3
 *  File: es6.c
 *  Date: 01/24/2006
 *  Author. R. Gioiosa
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/proc_fs.h>
#include <linux/kthread.h>

#define MODULE_VERS "1.0"
#define MODULE_NAME "LKH-ES6"
#define KNAME "lkh_timed"

#define DEBUG_ON

#ifdef DEBUG_ON
#define Dprintk(ARG...) printk(KERN_DEBUG MODULE_NAME ": " ARG)
#else
#define Dprintk(ARG...) 
#endif

#define N 32

static struct proc_dir_entry* lkh_pde;
static struct proc_dir_entry* entry;
static unsigned long time = 10;
static char buf[N];
struct semaphore time_sem;
pid_t tpid;
DECLARE_WAIT_QUEUE_HEAD(es6_wq);

static int es6_write(struct file* file,const char __user* buffer,unsigned long count,void* data)
{
  if(!count)
    {
      printk(KERN_ERR "Count is 0!!\n");
      return 0;
    }
  if(count>N)
    {
      Dprintk("Count is too large, set to %d\n",N-1);
      count = N-1;
    }
  if(copy_from_user(buf,buffer,count))
    {
      printk(KERN_ERR "Error while copying from user space!\n");
      return -EFAULT;
    }
  buf[count] = '\0';

  down(&time_sem);
  if(sscanf(buf,"%lu",&time) != 1)
    {
      up(&time_sem);
      printk(KERN_ERR "The value provided is not a valid number!\n");
      return -EFAULT;
    }
  up(&time_sem);
  Dprintk("writing time=%lu\n",time);

  return count;
}

static int es6_read(char* page,char** start,off_t off,int count,int* eof,void* data)
{
  int written = 0;

  down(&time_sem);
  written = sprintf(page,"var = %lu\n",time);
  up(&time_sem);
  *start = page+off;
  if(strlen(page) == written)
    *eof = 1;
  Dprintk("es6_read done.\n");
  return written-off;

}

int es6_ktime(void* data)
{
  unsigned int tsleep;
  
  down(&time_sem);
  tsleep = time;
  up(&time_sem);

  while(tsleep != 0)
    {
      Dprintk("Going to sleep...\n");
      sleep_on_timeout(&es6_wq,msecs_to_jiffies(tsleep*1000));
      Dprintk("Woken up!\n");

      down(&time_sem);
      tsleep = time;
      up(&time_sem);
    }
  Dprintk("Finished!\n");
  return 0;
}

static int __init es6_init(void)
{
  task_t* tsk = NULL;

  lkh_pde = proc_mkdir("lkh",NULL);
  if(!lkh_pde)
    {
      printk(KERN_ERR "%s: error creating proc_dir entry!\n",MODULE_NAME);
      goto err;
    }
  Dprintk("proc dir created\n");

  entry = create_proc_entry("foo",0444,lkh_pde);
  if(!entry)
    {
      printk(KERN_ERR "%s: error creating proc_entry!\n",MODULE_NAME);
      goto err_dir;
    }
  entry->data = NULL;
  entry->owner = THIS_MODULE;
  entry->read_proc = es6_read;
  entry->write_proc = es6_write;
  Dprintk("proc entry created\n");

  init_MUTEX(&time_sem);

  tsk = kthread_create(es6_ktime,NULL,KNAME);
  if(!tsk)
    {
      printk(KERN_ERR "%s: error creating kernel thread!\n",MODULE_NAME);
      goto err_entry;
    }
  tpid = tsk->pid;
  Dprintk("kernel thread %s created\n",KNAME);
  wake_up_process(tsk);
  Dprintk("kernel thread %s started\n",KNAME);

  Dprintk("module loaded\n");
  return 0;

 err_entry:
  remove_proc_entry("foo",lkh_pde);
  Dprintk("proc entry removed\n");

 err_dir:
  remove_proc_entry("lkh",NULL);
  Dprintk("proc dir removed\n");

 err:
  return -1;
}

static void __exit es6_exit(void)
{
  task_t* tsk;

  tsk = find_task_by_pid(tpid);
  if(tsk)
    {
      down(&time_sem);
      time = 0;
      up(&time_sem);
      kthread_stop(tsk);
      Dprintk("kernel thread stopped\n");
    }
  else
    Dprintk("kernel thread already stopped!\n");

  remove_proc_entry("foo",lkh_pde);
  Dprintk("proc entry removed\n");

  remove_proc_entry("lkh",NULL);
  Dprintk("proc dir removed\n");

  Dprintk("module unloaded\n");
}

module_init(es6_init);
module_exit(es6_exit);

MODULE_AUTHOR("Roberto Gioiosa");
MODULE_DESCRIPTION("Linux Kernel Hacking 06 - Es6");
MODULE_LICENSE("GPL");
MODULE_VERSION(MODULE_VERS);
